/*-
 * Copyright (c) 2019 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Implementation inspired by Varnish cache_gzip.c
 */

/* for strdup() */
#define _POSIX_C_SOURCE 200809L

#include "config.h"

#include <stdlib.h>
#include <string.h>
#include <brotli/encode.h>
#include <brotli/decode.h>

#include "cache/cache.h"
#include "cache/cache_filter.h"

#include "vcc_if.h"

#define VFAIL(ctx, fmt, ...) \
        VRT_fail((ctx), "vfp brotli failure: " fmt, __VA_ARGS__)

/* cf. varnishd default gzip_buffer */
#define DEFAULT_BUFSZ (32 * 1024 * 1024)

struct vbr_stream {
	const uint8_t		*next_in;
	uint8_t			*next_out;
	size_t			avail_in;
	size_t			avail_out;
	size_t			total_out;
};

enum vbr_which {ENC, DEC};

struct vbr {
	unsigned		magic;
#define VBR_MAGIC		0x467c387b
	struct vbr_stream	stream;
	union {
		BrotliEncoderState *enc;
		BrotliDecoderState *dec;
	}			state;
	uint8_t			*buf;
	size_t			bufsz;
	size_t			buflen;
	enum vbr_which		which;
};

struct vbr_settings {
	unsigned	magic;
#define VBR_SETTINGS_MAGIC 0xa61992aa
	VCL_BYTES	bufsz;
	enum vbr_which	which;
};

struct vmod_brotli_encoder {
	unsigned	magic;
#define VMOD_BROTLI_ENCODER_MAGIC 0x1490a42a
	struct vfp	*vfp;
};

struct custom_vfp_entry {
	unsigned	magic;
#define CUSTOM_VFP_MAGIC 0xfc88cb98
	VSLIST_ENTRY(custom_vfp_entry) list;
	struct vfp	*vfp;
};

VSLIST_HEAD(custom_vfp_head, custom_vfp_entry);

static struct vbr *
newVBR(enum vbr_which which)
{
	struct vbr *vbr;

	/* XXX inc counters */
	ALLOC_OBJ(vbr, VBR_MAGIC);
	if (vbr == NULL)
		return (NULL);
	if (which == ENC) {
		vbr->state.enc = BrotliEncoderCreateInstance(NULL, NULL, NULL);
		if (vbr->state.enc == NULL)
			return (NULL);
	}
	else {
		vbr->state.dec = BrotliDecoderCreateInstance(NULL, NULL, NULL);
		if (vbr->state.dec == NULL)
			return (NULL);
	}
	/* XXX configure the handle */
	vbr->which = which;
	return (vbr);
}

static void
destroy(struct vbr **vp)
{
	struct vbr *vbr;

	TAKE_OBJ_NOTNULL(vbr, vp, VBR_MAGIC);
	/* XXX log results: bytes in, bytes out */
	/* XXX IMO no finish op necessary here */
	if (vbr->which == ENC)
		BrotliEncoderDestroyInstance(vbr->state.enc);
	else
		BrotliDecoderDestroyInstance(vbr->state.dec);
	if (vbr->buf != NULL)
		free(vbr->buf);
	FREE_OBJ(vbr);
}

static int
getBuf(struct vbr *vbr, const struct vbr_settings *settings)
{
	CHECK_OBJ_NOTNULL(vbr, VBR_MAGIC);
	AZ(vbr->bufsz);
	AZ(vbr->buflen);
	AZ(vbr->buf);

	vbr->bufsz = settings->bufsz;
	vbr->buf = malloc(vbr->bufsz);
	if (vbr->buf == NULL) {
		vbr->bufsz = 0;
		return (-1);
	}
	return (0);
}

static const char *
decodeErrMsg(struct vbr *vbr)
{
	CHECK_OBJ_NOTNULL(vbr, VBR_MAGIC);
	assert(vbr->which == DEC);

	return (BrotliDecoderErrorString(
			BrotliDecoderGetErrorCode(vbr->state.dec)));
}

static void
setInputBuf(struct vbr *vbr, const void *ptr, ssize_t len)
{
	CHECK_OBJ_NOTNULL(vbr, VBR_MAGIC);

	AZ(vbr->stream.avail_in);
	vbr->stream.next_in = ptr;
	vbr->stream.avail_in = len;
}

static int
isInputBufEmpty(const struct vbr *vbr)
{
	CHECK_OBJ_NOTNULL(vbr, VBR_MAGIC);
	return (vbr->stream.avail_in == 0);
}

static void
setOutputBuf(struct vbr *vbr, const void *ptr, ssize_t len)
{
	CHECK_OBJ_NOTNULL(vbr, VBR_MAGIC);

	vbr->stream.next_out = TRUST_ME(ptr);
	vbr->stream.avail_out = len;
}

static BROTLI_BOOL
encode(struct vbr *vbr, ssize_t *dl, int finished)
{
	BROTLI_BOOL ret;
	const uint8_t *before;
	enum BrotliEncoderOperation op = BROTLI_OPERATION_PROCESS;

	CHECK_OBJ_NOTNULL(vbr, VBR_MAGIC);
	assert(vbr->which == ENC);
	before = vbr->stream.next_out;

	if (finished)
		op = BROTLI_OPERATION_FINISH;
	ret = BrotliEncoderCompressStream(vbr->state.enc, op,
					  &vbr->stream.avail_in,
					  &vbr->stream.next_in,
					  &vbr->stream.avail_out,
					  &vbr->stream.next_out,
					  &vbr->stream.total_out);
	*dl = (const uint8_t *)vbr->stream.next_out - before;
	return (ret);
}

static BrotliDecoderResult
decode(struct vbr *vbr, ssize_t *dl)
{
	BrotliDecoderResult ret;
	const uint8_t *before;

	CHECK_OBJ_NOTNULL(vbr, VBR_MAGIC);
	assert(vbr->which == DEC);
	before = vbr->stream.next_out;

	ret = BrotliDecoderDecompressStream(vbr->state.dec,
					  &vbr->stream.avail_in,
					  &vbr->stream.next_in,
					  &vbr->stream.avail_out,
					  &vbr->stream.next_out,
					  &vbr->stream.total_out);
	*dl = (const uint8_t *)vbr->stream.next_out - before;
	return (ret);
}

/* VFP interfaces */

static const struct vfp vfp_br, vfp_unbr;

/* init and fini methods for both br and unbr */

static enum vfp_status v_matchproto_(vfp_init_f)
vfp_br_init(struct vfp_ctx *ctx, struct vfp_entry *ent)
{
	struct vbr *vbr;
	const struct vbr_settings *settings;

	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(ent, VFP_ENTRY_MAGIC);
	AN(ent->vfp);
	CAST_OBJ_NOTNULL(settings, ent->vfp->priv1, VBR_SETTINGS_MAGIC);

	/*
	 * Ignore partial responses to range requests (as Varnish does for
	 * gzip and gunzip).
	 */
	if (http_GetStatus(ctx->resp) == 206)
	 	return (VFP_NULL);

	if (settings->which == ENC) {
		if (http_GetHdr(ctx->resp, H_Content_Encoding, NULL))
			return (VFP_NULL);
		vbr = newVBR(ENC);
	}
	else {
		if (!http_HdrIs(ctx->resp, H_Content_Encoding, "br"))
			return (VFP_NULL);
		vbr = newVBR(DEC);
	}
	if (vbr == NULL)
		return (VFP_ERROR);
	ent->priv1 = vbr;
	if (getBuf(vbr, settings))
		return (VFP_ERROR);
	setInputBuf(vbr, vbr->buf, 0);
	AZ(vbr->buflen);

	http_Unset(ctx->resp, H_Content_Encoding);
	http_Unset(ctx->resp, H_Content_Length);
	RFC2616_Weaken_Etag(ctx->resp);

	if (settings->which == ENC) {
		http_SetHeader(ctx->resp, "Content-Encoding: br");
		RFC2616_Vary_AE(ctx->resp);
	}

	return (VFP_OK);
}

static void v_matchproto_(vfp_fini_f)
vfp_br_fini(struct vfp_ctx *ctx, struct vfp_entry *ent)
{
	struct vbr *vbr;

	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(ent, VFP_ENTRY_MAGIC);

	if (ent->priv1 != NULL) {
		CAST_OBJ(vbr, ent->priv1, VBR_MAGIC);
		ent->priv1 = NULL;
		destroy(&vbr);
	}
}

/* pull for br (compression) */

static enum vfp_status v_matchproto_(vfp_pull_f)
vfp_br_pull(struct vfp_ctx *ctx, struct vfp_entry *ent, void *ptr,
	    ssize_t *lenp)
{
	struct vbr *vbr;
	ssize_t len, dl;
	enum vfp_status vp = VFP_ERROR;
	int finished = 0;
	BROTLI_BOOL done;

	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(ent, VFP_ENTRY_MAGIC);
	CAST_OBJ_NOTNULL(vbr, ent->priv1, VBR_MAGIC);
	assert(vbr->which == ENC);
	AN(ptr);
	AN(lenp);

	len = *lenp;
	*lenp = 0;
	setOutputBuf(vbr, ptr, len);

	do {
		if (isInputBufEmpty(vbr)) {
			len = vbr->bufsz;
			vp = VFP_Suck(ctx, vbr->buf, &len);
			if (vp == VFP_ERROR)
				break;
			if (vp == VFP_END)
				finished = 1;
			setInputBuf(vbr, vbr->buf, len);
		}
		if (!isInputBufEmpty(vbr) || finished) {
			done = encode(vbr, &dl, finished);
			if (done != BROTLI_TRUE)
				return (VFP_Error(ctx, "brotli encode failed"));
			if (dl > 0) {
				*lenp = dl;
				return (VFP_OK);
			}
		}
		AN(isInputBufEmpty(vbr));
	} while (!finished);

	if (done != BROTLI_TRUE)
		return (VFP_Error(ctx, "brotli encode failed"));
	return (VFP_END);
}

/* pull for unbr (decompression) */

static enum vfp_status v_matchproto_(vfp_pull_f)
vfp_unbr_pull(struct vfp_ctx *ctx, struct vfp_entry *ent, void *ptr,
	      ssize_t *lenp)
{
	struct vbr *vbr;
	ssize_t len, dl;
	enum vfp_status vp = VFP_ERROR;
	BrotliDecoderResult result = BROTLI_DECODER_RESULT_ERROR;

	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(ent, VFP_ENTRY_MAGIC);
	CAST_OBJ_NOTNULL(vbr, ent->priv1, VBR_MAGIC);
	assert(vbr->which == DEC);
	AN(ptr);
	AN(lenp);

	len = *lenp;
	*lenp = 0;
	setOutputBuf(vbr, ptr, len);

	do {
		if (isInputBufEmpty(vbr)) {
			len = vbr->bufsz;
			vp = VFP_Suck(ctx, vbr->buf, &len);
			if (vp == VFP_ERROR)
				break;
			setInputBuf(vbr, vbr->buf, len);
		}
		if (!isInputBufEmpty(vbr) || vp == VFP_END) {
			result = decode(vbr, &dl);
			if (result == BROTLI_DECODER_RESULT_SUCCESS &&
			    !isInputBufEmpty(vbr))
				return(VFP_Error(ctx,
						 "Junk after brotli data"));
			if (result == BROTLI_DECODER_RESULT_ERROR)
				return (VFP_Error(ctx,
						  "Invalid brotli data: %s",
						  decodeErrMsg(vbr)));
			if (dl > 0) {
				*lenp = dl;
				return (VFP_OK);
			}
		}
		AN(isInputBufEmpty(vbr));
	} while (result == BROTLI_DECODER_RESULT_NEEDS_MORE_INPUT);

	if (result != BROTLI_DECODER_RESULT_SUCCESS)
		return (VFP_Error(ctx, "Invalid brotli data at end: %s",
				  decodeErrMsg(vbr)));
	return (VFP_END);
}

static const struct vbr_settings default_encoder = {
	.magic = VBR_SETTINGS_MAGIC,
	.bufsz = DEFAULT_BUFSZ,
	.which = ENC,
};

static const struct vbr_settings default_decoder = {
	.magic = VBR_SETTINGS_MAGIC,
	.bufsz = DEFAULT_BUFSZ,
	.which = DEC,
};

static const struct vfp vfp_br = {
	.name = "br",
	.init = vfp_br_init,
	.pull = vfp_br_pull,
	.fini = vfp_br_fini,
	.priv1 = &default_encoder,
};

static const struct vfp vfp_unbr = {
	.name = "unbr",
	.init = vfp_br_init,
	.pull = vfp_unbr_pull,
	.fini = vfp_br_fini,
	.priv1 = &default_decoder,
};

/* Event function */

static struct custom_vfp_head *
init_priv_vcl(struct vmod_priv *priv)
{
	struct custom_vfp_head *vfph;

	AN(priv);
	if (priv->priv == NULL) {
		vfph = malloc(sizeof(*vfph));
		AN(vfph);
		priv->priv = vfph;
		VSLIST_INIT(vfph);
	}
	else
		vfph = priv->priv;
	return (vfph);
}

int
vmod_event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e e)
{
	struct custom_vfp_head *vfph;
	struct custom_vfp_entry *vfpe;
	struct vbr_settings *settings;

	ASSERT_CLI();
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(priv);

	vfph = init_priv_vcl(priv);

	switch(e) {
	case VCL_EVENT_LOAD:
		VRT_AddVFP(ctx, &vfp_br);
		VRT_AddVFP(ctx, &vfp_unbr);
		return (0);
	case VCL_EVENT_DISCARD:
		VRT_RemoveVFP(ctx, &vfp_br);
		VRT_RemoveVFP(ctx, &vfp_unbr);
		while (!VSLIST_EMPTY(vfph)) {
			vfpe = VSLIST_FIRST(vfph);
			CHECK_OBJ_NOTNULL(vfpe, CUSTOM_VFP_MAGIC);
			if (vfpe->vfp != NULL) {
				if (vfpe->vfp->priv1 != NULL) {
					CAST_OBJ(settings,
						 TRUST_ME(vfpe->vfp->priv1),
						 VBR_SETTINGS_MAGIC);
					FREE_OBJ(settings);
				}
				VRT_RemoveVFP(ctx, vfpe->vfp);
				free(vfpe->vfp);
			}
			VSLIST_REMOVE_HEAD(vfph, list);
			FREE_OBJ(vfpe);
		}
		free(vfph);
		return (0);
	case VCL_EVENT_WARM:
	case VCL_EVENT_COLD:
		return (0);
	default:
		WRONG("illegal event enum");
	}
	NEEDLESS(return (0));
}

/* Object encoder */

VCL_VOID
vmod_encoder__init(VRT_CTX, struct vmod_brotli_encoder **encp,
		   const char *vcl_name, struct vmod_priv *priv,
		   VCL_STRING filter_name, VCL_BYTES bufsz)
{
	struct vmod_brotli_encoder *enc;
	struct vfp *vfp;
	struct vbr_settings *settings;
	struct custom_vfp_head *vfph;
	struct custom_vfp_entry *vfpe;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(encp);
	AZ(*encp);
	AN(vcl_name);
	AN(priv);
	AN(filter_name);
	assert(bufsz >= 0);

	if (*filter_name == '\0') {
		VFAIL(ctx, "new %s: filter name must be non-empty", vcl_name);
		return;
	}
	if (strcmp(filter_name, "br") == 0 || strcmp(filter_name, "unbr") == 0
	    || strcmp(filter_name, "esi") == 0
	    || strcmp(filter_name, "esi_gzip") == 0
	    || strcmp(filter_name, "gunzip") == 0
	    || strcmp(filter_name, "gzip") == 0
	    || strcmp(filter_name, "testgunzip") == 0) {
		VFAIL(ctx,
		      "new %s: filter name %s already in use by another VFP",
		      vcl_name, filter_name);
		return;
	}
	if (bufsz == 0) {
		VFAIL(ctx, "new %s: buffer size must be > 0b", vcl_name);
		return;
	}

	errno = 0;
	ALLOC_OBJ(enc, VMOD_BROTLI_ENCODER_MAGIC);
	if (enc == NULL) {
		VFAIL(ctx, "new %s: cannot allocate space for object: %s",
		      vcl_name, strerror(errno));
		return;
	}

	errno = 0;
	vfp = malloc(sizeof(*vfp));
	if (vfp == NULL) {
		VFAIL(ctx, "new %s: cannot allocate space for VFP: %s",
		      vcl_name, strerror(errno));
		return;
	}

	errno = 0;
	ALLOC_OBJ(settings, VBR_SETTINGS_MAGIC);
	if (vfp == NULL) {
		VFAIL(ctx, "new %s: cannot allocate space for settings: %s",
		      vcl_name, strerror(errno));
		return;
	}

	errno = 0;
	ALLOC_OBJ(vfpe, CUSTOM_VFP_MAGIC);
	if (vfpe == NULL) {
		VFAIL(ctx, "new %s: cannot allocate space VFP entry: %s",
		      vcl_name, strerror(errno));
		return;
	}

	settings->bufsz = bufsz;
	settings->which = ENC;

	vfp->name = strdup(filter_name);
	vfp->init = vfp_br_init;
	vfp->pull = vfp_br_pull;
	vfp->fini = vfp_br_fini;
	vfp->priv1 = settings;
	enc->vfp = vfp;
	VRT_AddVFP(ctx, vfp);

	vfph = init_priv_vcl(priv);
	vfpe->vfp = vfp;
	VSLIST_INSERT_HEAD(vfph, vfpe, list);
}

/*
 * The settings and custom VFP objects and PRIV_VCL list entry are freed
 * on the DISCARD event, because we need a VRT_CTX to call
 * VRT_RemoveVFP().
 */
VCL_VOID
vmod_encoder__fini(struct vmod_brotli_encoder **encp)
{
	struct vmod_brotli_encoder *enc;

	if (encp == NULL || *encp == NULL)
		return;
	CHECK_OBJ(*encp, VMOD_BROTLI_ENCODER_MAGIC);
	enc = *encp;
	*encp = NULL;
	FREE_OBJ(enc);
}

/* Version functions */

static VCL_STRING
brotli_version(VRT_CTX, uint32_t version)
{
	VCL_STRING vers;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	unsigned major = version >> 24;
	unsigned minor = (version >> 12) & 0xFFF;
	unsigned patch = version & 0xFFF;
	if ((vers = WS_Printf(ctx->ws, "%u.%u.%u", major, minor, patch))
	    == NULL) {
		VSLb(ctx->vsl, SLT_VCL_Error, "vfp brotli error: "
		     "insufficient workspace for version string");
		return ("(insufficient workspace for version string)");
	}
	return vers;
}

VCL_STRING
vmod_encoder_version(VRT_CTX)
{
	return (brotli_version(ctx, BrotliEncoderVersion()));
}

VCL_STRING
vmod_decoder_version(VRT_CTX)
{
	return (brotli_version(ctx, BrotliDecoderVersion()));
}

VCL_STRING
vmod_version(VRT_CTX)
{
	(void) ctx;
	return VERSION;
}
